#!/usr/local/bin/perl

# Parse XDOC (Xerox Imaging Systems OCR format) 
# Jim Davis 15 Sept 93
# modified 10 Nov 93 to parse graphic zones
# modified 11 Nov 93 to include zone number in XD_beginline
# modified 12 Nov 93 to include nspaces and ntabs in XD_whitespace
#  and to handle tabs in XD_whitespace
# modified 24 Feb 94 to get rid of all control characters.  Only
#  newline survives.  
# modified 27 Apr 94.  Added position args to whitespace, beginline;
#   added XD_beginpage; added code to ignore text before the "a" command.
# 2 May 94 - unify pass1 and pass2
# 6 May 94 - define XIS extended characters ($XIS_L_SINGLE_QUOTE etc)
#  and translate them to sane characters.  Handle the "default font" arg in start-of-line;
#  keep track of current font and don't switch if no change.
# 4 May 95 - partial mods for XDOC 10.0
# 11 June 95 - modify calling sequence for XD_beginline and XD_endofline to
#  pass new paramters $baseline and $new_zone

# This reads the file, parses the markup, and calls routines
# (which you define) as it encounters text, font shifts, etc
# in the text.  Default definitions are provided which simply
# print the text in plain ASCII.  If you want to do something
# fancier like generate font shifts, you can define the routines.

# Pass 1 - read file, extract font info
# so that when we read the file next time we'll know the fonts
# Pass 2 - read the file again, this time we know the font and zone info.

# The documentation says the version is XDOC9.0 but in fact it's XDOC.9.0
# note that it's also in quotes.

$xdoc_version9 = "\"XDOC.9.0\"";
$xdoc_version10 = "\"XDOC.10.0\"";
$supported_versions = "9, 10";

sub parse_xdoc_file {
    local ($file) = @_;
    local (@fonts, @zones);
    @fonts = ();
    @zones = ();

    # various temps used inside handlers for commands
    local (@zone_left, @zone_top, @zone_right, @zone_bottom);
    local (@args);
    local ($advances, $baseline, $bottom, $c, $char, $confidence, $distance,
	   $face, $fixed, $fn, $font,
	   $gap, $hard, $height, $height_descender, $height_lc, $height_uc,
	   $indent, $interval, $left, $length, $n, $name, $nspaces, $ntabs, $nzones, 
	   $obsolete, $orientation, $point_size,
	   $rest, $res1, $res2, $right,
	   $serif, $sibling, $style, $tab, $top, $untilt, $version, $width, $xmid, $ymid, $zone);

    # state
    local ($superscript) = 0;
    local ($subscript) = 0;
    local ($underline) = 0;
    local ($font) = 0;
    local ($text_begun) = 0;	# ignore text before document.
    local ($line_counter) = 0;
    local ($current_zone) = 0;
    local ($current_font) = -1;
    local ($new_zone) = 0;      # for new para/map files, 5/95 ESS.

    foreach $pass  (1, 2) {

	if (! open (IN, $file)) {
	    die "Can't open $file: $!\nStopped";}
	&XD_begin_of_file ($pass, $file, *fonts, *zones);

	$text_begun = 0;
	$line_counter = 0;
	$current_font = -1;

	while (<IN>) {
	    $line_counter++;

	    while ($_ ne "") {
		if (/([^\[]*)\[(.)(.*)/) {
		    if ($1 ne "")   {&XD_text ($pass, $1);}
		    $c = $2;
		    $rest = $3;

		    # Dispatch on command.
		    if ($c =~ /[A-Z]/) {
			# Upper case letter, no operands
			if ($c eq "A") {
			    # end of table
			    &xdoc_error ("Table markup not supported.");}
			elsif ($c eq "B") {
			    $subscript = $subscript?0:1;
			    &XD_subscript($pass, $subscript);}			
			elsif ($c eq "E") {
			    # unrecognized character
			    &XD_unrecognized ($pass);}
			elsif ($c eq "H") {
			    &XD_softhyphen($pass);}
			elsif ($c eq "M") {
			    # numeric word
			}
			elsif ($c eq "Q") {
			    # questionable character
			}
			elsif ($c eq "S") {
			    # toggle superscript
			    $superscript= $superscript?0:1;
			    &XD_superscript($pass, $superscript);}
			elsif ($c eq "U") {
			    # toggle underline
			    $underline= $underline?0:1;
			    &XD_underline($pass, underline);}
			elsif ($c eq "Z") {
			    # end of document
			}
			else {
			    # unknown character
			    &xdoc_error ("Unrecognized escape code %s ignored.", $c);
			}
		    }		# end uppercase dispatch

		    elsif ($c =~ /[a-z]/) {
			# lowercase, get operands
			$rest =~ /;([^\]]+)\](.*)$/;
			@args = split (/;/, $1);
			$rest = $2;

			# Dispatch on operand
			if ($c eq "a") {
			    ($version) = @args;
			    if ($version eq $xdoc_version9) {
			    }
			    elsif ($version eq $xdoc_version10) {
			    }
			    else {
				&xdoc_error ("XDOC version %s is not supported.  Should be %s.",
					     $version, $supported_versions);}
			    $text_begun = 1;}
			#"b" bounding box for next word
			elsif ($c eq "c") {
			    # change font
			    ($font) = @args;
			    if ($pass == 2) {
				if ($current_font != $font) {
				    &switch_to_font ($font);
				    $current_font = $font;}}}
			elsif ($c eq "d") {
			    # internal doc name
			}
			elsif ($c eq "e") {
			    # change region
			}
			elsif ($c eq "f") {
			    if ($pass == 1) {
				($font, $name, $face, $serif, $width, $fixed,
				 $height_uc, $height_descender, $height_lc,
				 $point_size) = @args;
				$fonts[$font] = join(";",$face,$serif,$fixed,$point_size,
						     $height_uc,$height_descender,$height_lc);}}
			elsif ($c eq "g") {
			    # pass 1
			}
			elsif ($c eq "h") {
			    # nonprinting whitespace
			    ($left, $width, $nspaces, $ntabs) = @args;
			    &XD_whitespace ($pass, $left, $width, $nspaces, $ntabs);}

			elsif ($c eq "i") {
			    $arg = $args[0];
			    &xdoc_error ("Literal sequence (%s) not supported.", $arg);}

			elsif ($c eq "j") {
			    # begin table
			    ($id, $left, $top, $right, $bottom, $cols, $rows, $position, @col_coords) 
				= @args;
			    &xdoc_error ("Table markup is not supported.");}
			elsif ($c eq "k") {
			    # Section change (V 10)
			    ($type, $ncols, $picture_id, $extra, @coordinates)
				= @args;
			}
			elsif ($c eq "l") {
			    # leader
			    ($char, $left, $width, $nspaces, $ntabs) = @args;
			}
			elsif ($c eq "n") {
			    # begin new cell in a table
			    &xdoc_error ("Table markup not supported.");}
			elsif ($c eq "o") {
			    # end row of table
			    &xdoc_error ("Table markup not supported.");}
			elsif ($c eq "p") {
			    # page start
			    if ($version eq $xdoc_version9) {
				($n, $orientation, $nzones, $untilt,
				 $left, $top, $width, $height) = @args;}

			    elsif ($version eq $xdoc_version10) {
				($n, $orientation, $columns, $mode,
				 $skew, $untilt, $left, $top, $width, $height) = @args;}
			    &XD_beginpage ($pass, $left, $top, $width, $height);}
			elsif ($c eq "q") {
			    # confidence
			    ($confidence) = @args;
			}
			elsif ($c eq "r") {
			    # ruling
			    ($xmid, $ymid, $orientation, $length,
			     $style, $thickness, $interval) = @args;
			} 
			elsif ($c eq "s") {
				# start of text line
			    if ($version eq $xdoc_version9) {
				($zone, $left, $distance, $baseline,
				 $style, $font, $indent, $tab) = @args;}
			    elsif ($version eq $xdoc_version10) {
				($zone, $left, $distance, $baseline,
				 $style, $font, $indent, $nspaces, $tab) = @args;}
			    if($current_zone ne $zone) { 
                                $new_zone = 1; }
                            else {
                                $new_zone = 0; }

			    $current_zone = $zone;
			    if ($pass == 1) {
				if ($zone_left [$zone] == 0 || 
				    $left < $zone_left [$zone]) {
				    $zone_left[$zone] = $left;} }
			    &XD_beginline ($pass, style, $zone, $left, $distance, $baseline, $new_zone);
			    if ($pass == 2) {
				if ($current_font != $font) {
				    &switch_to_font ($font);
				    $current_font = $font;}}}
			elsif ($c eq "t") {
			    # text zone descriptor
			    if ($version eq $xdoc_version9) {
				($zone, $top, $height, $obsolete,
				 $sibling, $res1, $res2, $name) = @args;}
			    elsif ($version eq $xdoc_version10) {
				($zone, $order, $top, $height, $region,
				 $prefix, $suffix, $name,
				 $left1, $top1, $right1, $bottom1,
				 $language) = @args;}
			    if ($pass == 1) {
				$zone_top[$zone] = $top;
				$zone_bottom [$zone] = ($top + $height);}}
			elsif ($c eq "u") {
			    # dropped capital
			    &xdoc_error ("Dropped capital not supported.");}
			elsif ($c eq "v") {
			    # lexical class
			}
			elsif ($c eq "w") {
			    # word confidence
			    ($confidence) = @args;
			}
			elsif ($c eq "x") {
			    ($zone, $compression, $top, $left, $right, $bottom) = @args;
			    &XD_declare_graphic_zone ($pass, $zone, $compression, 
						      $left, $top, $right-$left, $bottom-$top);
			}
			elsif ($c eq "y") {
			    # text line information
			    ($right, $gap, $baseline, $advances, $hard) = @args;
			    if ($pass == 1) {
				if ($right > $zone_right [$current_zone]) {
				    $zone_right[$current_zone] = $right;} }
			    &XD_endofline ($pass, ($hard eq "H"), $right, $gap, $baseline, $new_zone);}
			else {
			    &xdoc_error ("Unrecognized escape code %s ignored.", $c);}
		    }		# end lowercase dispatch
		    elsif ($c eq "[") {
			&XD_text ($pass, "[");}
		    else {
			&xdoc_error ("Unrecognized escape code %s, ignoring rest of line.",
				     $c);
			last;}
		    $_ = $rest;
		}
		## no command, rest is text
		else {
		    if ($_ ne "" && $text_begun) {&XD_text ($pass, $_);}
		    last;}
	    }}
	close (IN);

	if ($pass == 1) {
	    local ($i);
	    for ($i = 0 ; $i <= $#zone_right; $i++) {
		$zones[$i] = join (":", $zone_left[$i], $zone_top[$i],
				   $zone_right[$i], $zone_bottom [$i]);}}

	&XD_end_of_file ($pass, $file);
    }
}

sub switch_to_font {
    local ($font) = @_;
    if ($font == 0) {
	# ScanWorX uses this number for those fonts for which there are
	# too few characters to reliably estimate size.
    }
    else {
	($face, $serif, $fixed, $point_size, $height_uc, $height_descender, $height_lc) = 
	    split (/;/, $fonts[$font]);
	&XD_font ($pass, $face, $serif, $fixed, $point_size,
		  $height_uc, $height_descender, $height_lc);}}


# Error messages are printed only during pass 1
# global read $pass, $file, $line_counter
sub xdoc_error {
    if ($pass == 1) {
	printf STDERR @_;
	print STDERR "\n  $file line $line_counter\n";}}



## default bindings for the routines, useful to print out plain ascii
# For all routines (except where noted) the first arg is always the pass
# number. 

# Called at the beginning of processing.
# no pass number argument, it would not make sense.
sub XD_begin_document {
    local ($nfiles) = @_;
}


# Only called on Pass 2, cause font info is not available Pass 1
$font_fixed = 0;
$font_size = 12;
$font_xheight = 20;

# Compare the width of the whitespace to the current x-height
# (the height of an x character).  For a fixed width font the space
# must be >= .75 the x-height.  For a variable width font the cutoff
# is much lower, because it's better to insert a spurious space than
# to throw one away.  The value of .35 comes because I saw at least
# one case where the ratio was .39  It may be that I will have to
# consider kerning.  This low value was an f followed by a t.

sub XD_font {
    local ($pass, $face, $serif, $fixed, $point_size, 
	   $height_uc, $height_descender, $height_lc) = @_;

    $font_size = $point_size;
    $font_xheight = $height_lc;

    # F means fixed, V is variable.
    # B is Bold, R is Roman, I Italic
    if ($fixed eq "F") {
	$font_fixed = 1;}
    elsif ($fixed eq "V") {
	$font_fixed = 0;
    }}

sub XD_subscript {
    local ($pass, $on) = @_;
}

sub XD_superscript {
    local ($pass, $on) = @_;
}

sub XD_underline {
    local ($pass, $on) = @_;
}


$XIS_L_SINGLE_QUOTE = "\201";
$XIS_R_SINGLE_QUOTE = "\202";
$XIS_L_DOUBLE_QUOTE = "\203";
$XIS_R_DOUBLE_QUOTE = "\204";
$XIS_EM_DASH = "\205";
$XIS_BULLET = "\206";


sub XD_text {
    local ($pass, $_) = @_;
    if ($pass == 2) {
	s/$XIS_L_SINGLE_QUOTE/`/g;
	s/$XIS_R_SINGLE_QUOTE/'/g;
	s/$XIS_L_DOUBLE_QUOTE/"/g;
        s/$XIS_R_DOUBLE_QUOTE/"/g;
        s/$XIS_EM_DASH/--H/g;
        s/$XIS_BULLET/o\010+/g;
	tr/\000-\007\013-\037\177-\255//d;
	print;}}


sub XD_whitespace {
    local ($pass, $left, $width, $nspaces, $ntabs) = @_;
    if ($pass == 2) {
	if ($ntabs ne "") {
	    local ($i);
	    for ($i=0,$i<$ntabs,$i++) {print "\t";}
#	for ($i=0,$i<$ntabs,$i++) {print " | ";}
	}
	else {
	    if ($width >= $font_xheight * ($font_fixed?.75:.35)) {
		print " ";}}}}

# style will be "c" for centered text or "p" for paragraph.
sub XD_beginline {
    local ($pass, $style, $zone, $left, $distance, $baseline) = @_;}

sub XD_endofline {
    local ($pass, $hard, $right, $gap) = @_;
    if ($pass == 2) {
	print "\n";}}

$unrecognized_character = "?";

sub XD_unrecognized {
    if ($pass == 2) {
	print $unrecognized_character;}}

sub XD_softhyphen {
    if ($pass == 2) {
	print "-";}}

sub XD_declare_graphic_zone {
    local ($pass, $zone, $compression, $left, $top, $width, $height) = @_;
}

sub XD_beginpage {
    local ($pass, $tilt, $left, $top, $width, $height) = @_;
}

sub XD_begin_of_file {
    local ($pass, $file) = @_;
}

sub XD_end_of_file {
    local ($pass, $file) = @_;
}
